#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic proxy to access any DFP web service."""

__author__ = 'api.jdilallo@gmail.com (Joseph DiLallo)'

import time

from adspygoogle import SOAPpy
from adspygoogle.common import Utils
from adspygoogle.common.Errors import Error
from adspygoogle.common.Errors import ValidationError
from adspygoogle.common.GenericApiService import GenericApiService
from adspygoogle.common.GenericApiService import MethodInfoKeys
from adspygoogle.dfp import AUTH_TOKEN_EXPIRE
from adspygoogle.dfp import AUTH_TOKEN_SERVICE
from adspygoogle.dfp import LIB_SIG
from adspygoogle.dfp import LIB_URL
from adspygoogle.dfp.DfpErrors import DfpApiError
from adspygoogle.dfp.DfpErrors import DfpError
from adspygoogle.dfp.DfpErrors import ERRORS
from adspygoogle.dfp.DfpSoapBuffer import DfpSoapBuffer


class GenericDfpService(GenericApiService):

  """Wrapper for any DFP web service."""

  # The _IGNORED_HEADER_VALUES are keys in the headers dictionary passed into
  # this class' constuctor which should NOT be packed into the SOAP header.
  _IGNORED_HEADER_VALUES = ('authToken', 'email', 'password',
                            'oauth2credentials')
  # The _WRAP_LISTS constant indicates that DFP services do not need to wrap
  # lists in an extra layer of XML element tags.
  _WRAP_LISTS = False
  # The _BUFFER_CLASS is the subclass of SoapBuffer that should be used to track
  # all SOAP interactions
  _BUFFER_CLASS = DfpSoapBuffer

  def __init__(self, headers, config, op_config, lock, logger, service_name):
    """Inits GenericDfpService.

    Args:
      headers: dict Dictionary object with populated authentication
               credentials.
      config: dict Dictionary object with populated configuration values.
      op_config: dict Dictionary object with additional configuration values for
                 this operation.
      lock: thread.lock Thread lock to use to synchronize requests.
      logger: Logger Instance of Logger to use for logging.
      service_name: string The name of this service.
    """
    service_url = '/'.join([op_config['server'], 'apis/ads/publisher',
                            op_config['version'], service_name])
    namespace = '/'.join(['https://www.google.com/apis/ads/publisher',
                          op_config['version']])
    namespace_extractor = _DetermineNamespacePrefix

    super(GenericDfpService, self).__init__(
        headers, config, op_config, lock, logger, service_name, service_url,
        GenericDfpService._WRAP_LISTS, GenericDfpService._BUFFER_CLASS,
        namespace, namespace_extractor)

    # DFP-specific changes to the SOAPpy.WSDL.Proxy
    methodattrs = {
        'xmlns:dfp': self._namespace,
        'xmlns': self._namespace
    }
    self._soappyservice.soapproxy.methodattrs = methodattrs

  def _SetHeaders(self):
    """Sets the SOAP headers for this service's requests."""
    now = time.time()
    if ((('authToken' not in self._headers and
          'auth_token_epoch' not in self._config) or
         int(now - self._config['auth_token_epoch']) >= AUTH_TOKEN_EXPIRE) and
        not self._headers.get('oauth2credentials')):
      if ('email' not in self._headers or not self._headers['email'] or
          'password' not in self._headers or not self._headers['password']):
        raise ValidationError('Required authentication headers, \'email\' and '
                              '\'password\', are missing. Unable to regenerate '
                              'authentication token.')
      self._headers['authToken'] = Utils.GetAuthToken(
          self._headers['email'], self._headers['password'], AUTH_TOKEN_SERVICE,
          LIB_SIG, self._config['proxy'])
      self._config['auth_token_epoch'] = time.time()

    # Apply headers to the SOAPpy service.
    soap_headers = SOAPpy.Types.headerType(attrs={'xmlns': self._namespace})
    request_header_data = {}
    if 'authToken' in self._headers:
      authentication_block = SOAPpy.Types.structType(
          data={'token': self._headers['authToken']},
          name='authentication', typed=0,
          attrs={(SOAPpy.NS.XSI3, 'type'): 'ClientLogin'})
      request_header_data['authentication'] = authentication_block
    for key in self._headers:
      if (key in GenericDfpService._IGNORED_HEADER_VALUES or
          not self._headers[key]):
        continue
      value = self._headers[key]
      if key == 'applicationName':
        value = ''.join([value, LIB_SIG])
      request_header_data[key] = SOAPpy.Types.stringType(value)
    request_header = SOAPpy.Types.structType(
        data=request_header_data, name='RequestHeader', typed=0)
    soap_headers.RequestHeader = request_header
    if 'authToken' in self._headers:
      soap_headers.RequestHeader._keyord = ['applicationName', 'authentication']
    self._soappyservice.soapproxy.header = soap_headers

  def _GetMethodInfo(self, method_name):
    """Pulls all of the relevant data about a method from a SOAPpy service.

    The return dictionary has two keys, MethodInfoKeys.INPUTS and
    MethodInfoKeys.OUTPUTS. Each of these keys has a list value. These lists
    contain a dictionary of information on the input/output parameter list, in
    order.

    Args:
      method_name: string The name of the method to pull information for.
    Returns:
      dict A dictionary containing information about a SOAP method.
    """
    rval = {}
    rval[MethodInfoKeys.INPUTS] = []
    for i in range(len(self._soappyservice.wsdl.types[
        self._namespace].elements[method_name].content.content.content)):
      param_attributes = self._soappyservice.wsdl.types[
          self._namespace].elements[method_name].content.content.content[
              i].attributes
      inparam = {
          MethodInfoKeys.ELEMENT_NAME: param_attributes['name'],
          MethodInfoKeys.NS: param_attributes['type'].getTargetNamespace(),
          MethodInfoKeys.TYPE: param_attributes['type'].getName(),
          MethodInfoKeys.MAX_OCCURS: param_attributes['maxOccurs']
      }
      rval[MethodInfoKeys.INPUTS].append(inparam)

    rval[MethodInfoKeys.OUTPUTS] = []
    for i in range(len(self._soappyservice.wsdl.types[
        self._namespace].elements[
            method_name + 'Response'].content.content.content)):
      param_attributes = self._soappyservice.wsdl.types[
          self._namespace].elements[
              method_name + 'Response'].content.content.content[i].attributes
      outparam = {
          MethodInfoKeys.ELEMENT_NAME: param_attributes['name'],
          MethodInfoKeys.NS: param_attributes['type'].getTargetNamespace(),
          MethodInfoKeys.TYPE: param_attributes['type'].getName(),
          MethodInfoKeys.MAX_OCCURS: param_attributes['maxOccurs']
      }
      rval[MethodInfoKeys.OUTPUTS].append(outparam)
    return rval

  def _HandleLogsAndErrors(self, buf, start_time, stop_time, error=None):
    """Manage SOAP XML message.

    Args:
      buf: SoapBuffer SOAP buffer.
      start_time: str Time before service call was invoked.
      stop_time: str Time after service call was invoked.
      [optional]
      error: dict Error, if any.
    """
    if error is None:
      error = {}
    try:
      handlers = self.__GetLogHandlers(buf)

      fault = super(GenericDfpService, self)._ManageSoap(
          buf, handlers, LIB_URL, start_time, stop_time, error)
      if fault:
        # Raise a specific error, subclass of DfpApiError.
        if 'detail' in fault:
          if 'code' in fault['detail']:
            code = int(fault['detail']['code'])
            if code in ERRORS: raise ERRORS[code](fault)
          elif 'errors' in fault['detail']:
            error_type = fault['detail']['errors'][0]['type']
            if error_type in ERRORS: raise ERRORS[str(error_type)](fault)

        if isinstance(fault, basestring):
          raise DfpError(fault)
        elif isinstance(fault, dict):
          raise DfpApiError(fault)
    except DfpApiError, e:
      raise e
    except DfpError, e:
      raise e
    except Error, e:
      if error: e = error
      raise Error(e)

  def __GetLogHandlers(self, buf):
    """Gets a list of log handlers for the DFP library.

    Args:
      buf: SoapBuffer SOAP buffer from which calls are retrieved for logging.

    Returns:
      list Log handlers for the DFP library.
    """
    return [
        {
            'tag': 'xml_log',
            'name': 'soap_xml',
            'data': ''
        },
        {
            'tag': 'request_log',
            'name': 'request_info',
            'data': str('host=%s service=%s method=%s responseTime=%s '
                        'requestId=%s'
                        % (Utils.GetNetLocFromUrl(self._service_url),
                           self._service_name, buf.GetCallName(),
                           buf.GetCallResponseTime(), buf.GetCallRequestId()))
        },
        {
            'tag': '',
            'name': 'dfp_api_lib',
            'data': ''
        }
    ]


def _DetermineNamespacePrefix(url):
  """Returns the SOAP prefix to use for definitions within the given namespace.

  Args:
    url: string The URL of the namespace.

  Returns:
    string The SOAP namespace prefix to use for the given namespace.
  """
  return 'dfp:'



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handy utility functions."""

__author__ = 'api.jdilallo@gmail.com (Joseph DiLallo)'

import os

from adspygoogle.common import Utils
from adspygoogle.common.Errors import ValidationError
from adspygoogle.dfa import LIB_HOME


def GetErrorCodes():
  """Get a list of available error codes.

  Returns:
    list Available error codes.
  """
  return Utils.GetDataFromCsvFile(os.path.join(LIB_HOME, 'data',
                                               'error_codes.csv'))


def GetAdXsiTypes():
  """Gets a dictionary of ad xsi_types indexed by their 'typeId' values.

  Returns:
    dict Ad xsi_types indexed by 'typeId'.
  """
  return Utils.GetDictFromCsvFile(os.path.join(LIB_HOME, 'data',
                                               'ad_types.csv'))


def GetCreativeXsiTypes():
  """Gets a dictionary of creative xsi_types indexed by their 'typeId' values.

  Returns:
    dict Creative xsi_types indexed by 'typeId'.
  """
  return Utils.GetDictFromCsvFile(os.path.join(LIB_HOME, 'data',
                                               'creative_types.csv'))


def AssignAdXsi(ad):
  """Assigns an xsi type to an Ad object if one is not present.

  Args:
    ad: dict Ad object.

  Raises:
    ValidationError: The ad object is lacking enough information to determine
                     what concrete class it represents.
  """
  if 'xsi_type' not in ad:
    if 'typeId' in ad:
      ad['xsi_type'] = GetAdXsiTypes()[ad['typeId']]
    else:
      raise ValidationError('The type of the ad is missing.')


def AssignCreativeXsi(creative):
  """Assigns an xsi type to a creative object if one is not present.

  Args:
    creative: dict CreativeBase object.

  Raises:
    ValidationError: The creative object is lacking enough information to
                     determine what concrete class it represents.
  """
  if 'xsi_type' not in creative:
    if 'typeId' in creative:
      creative['xsi_type'] = GetCreativeXsiTypes()[creative['typeId']]
    else:
      raise ValidationError('The type of the creative is missing.')


def DetermineServiceFromUrl(url):
  """Takes a DFA service's URL and returns the service name.

  Args:
    url: string The DFA service's URL.

  Returns:
    string The name of the service this URL points to.
  """
  return url.split('/')[-1]



#!/usr/bin/python
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for handling logging."""

__author__ = 'api.arogal@gmail.com (Adam Rogal)'

import logging
import os
import sys


class Logger(object):

  """Responsible for impementing logging.

  Allows to write to an external log or a console. There are 4 types of
  handlers denoted by the handler constants: NONE, FILE, CONSOLE, and
  FILE_AND_CONSOLE. There are also 6 verbosity levels defined by the constants:
  NOTSET, DEBUG, INFO, WARNING, ERROR, and CRITICAL. These constants are
  derived from the original logging module with CRITCAL being the highest
  importance.

  This class is a wrapper for the standard logging module.
  """

  # Handler constants.
  NONE = 0
  FILE = 1
  CONSOLE = 2
  FILE_AND_CONSOLE = 3

  # Level constants.
  CRITICAL = logging.CRITICAL
  ERROR = logging.ERROR
  WARNING = logging.WARNING
  INFO = logging.INFO
  DEBUG = logging.DEBUG
  NOTSET = logging.NOTSET

  def __init__(self, lib_sig, log_path=os.path.join(os.getcwd(), 'logs')):
    """Inits Logger.

    Args:
      lib_sig: str Signature of the client library.
      [optional]
      log_path: str Absolute or relative path to the logs directory.
    """
    self.__lib_sig = lib_sig
    self.__log_path = log_path
    self.__log_table = {}

  def __CreateLog(self, log_name, log_level=NOTSET, log_handler=FILE,
                  stream=sys.stderr):
    """Creates the log used for logging.

    Args:
      log_name: str Name of the log. If the log is handled by an external
                file, this will be the file name.
      [optional]
      log_level: str Level of the log. Should be one of CRITICAL, ERROR,
                 WARNING, INFO, DEBUG, or NOTSET.
      log_handler: int Type of log handler. Should be one of NONE, FILE,
                   CONSOLE, or FILE_AND_CONSOLE.
      stream: file Stream to send data into.
    """
    logger = logging.getLogger(log_name)

    # Update log level to reflect changes. If a higher log level is given
    # the logger should raise it's boundary.
    if log_level < logger.level or logger.level == logging.NOTSET:
      logger.setLevel(log_level)

    if (log_name in self.__log_table and
        self.__log_table[log_name] == Logger.FILE_AND_CONSOLE):
      # Don't add any more handlers.
      return

    # Create an entry for log name.
    if log_name not in self.__log_table:
      self.__log_table[log_name] = Logger.NONE

    if log_handler != Logger.NONE:
      fmt = ('[%(asctime)s::%(levelname)s::' + self.__lib_sig +
             '] %(message)s')
      # Add FILE handler if needed.
      if (log_handler == Logger.FILE or
          log_handler == Logger.FILE_AND_CONSOLE and
          self.__log_table[log_name] != Logger.FILE):
        if not os.path.exists(self.__log_path):
          os.makedirs(self.__log_path)
        fh = logging.FileHandler(os.path.join(self.__log_path,
                                              '%s.log' % log_name))
        fh.setLevel(log_level)
        fh.setFormatter(logging.Formatter(fmt))
        logger.addHandler(fh)
        # Binary arithmetic to yield updated handler.
        self.__log_table[log_name] = self.__log_table[log_name] + Logger.FILE

      # Add CONSOLE handler if needed.
      if (log_handler == Logger.CONSOLE or
          log_handler == Logger.FILE_AND_CONSOLE and
          self.__log_table[log_name] != Logger.CONSOLE):
        ch = logging.StreamHandler(stream)
        ch.setLevel(log_level)
        ch.setFormatter(logging.Formatter(fmt))
        logger.addHandler(ch)
        # Binary arithmetic to yield updated handler.
        self.__log_table[log_name] = self.__log_table[log_name] + Logger.CONSOLE

  def Log(self, log_name, message, log_level=NOTSET, log_handler=FILE):
    """Log message to an external file.

    Args:
      log_name: str Name of the log. If the log is handled by an external
                file, this will be the file name appended by log.
      message: str Message to log.
      [optional]
      log_level: int Level of importance of the current message. Not
                 supplying this parameter will cause the logger to log at the
                 lowest level of its handlers. If a handler has cut-off higher
                 than this lowest level, the message will be ignored by that
                 handler.
      log_handler: int Type of log handler. Should be one of NONE, FILE,
                   CONSOLE, or FILE_AND_CONSOLE.
    """
    logger = logging.getLogger(log_name)

    # Instantiate handlers for logger with default values if none exists.
    if not logger.handlers:
      self.__CreateLog(log_name, log_level, log_handler)

    if log_level == Logger.NOTSET:
      logger.log(logger.getEffectiveLevel(), message)
    else:
      logger.log(log_level, message)



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation functions."""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

from adspygoogle.common.Errors import ValidationError


def ValidateServer(server, version):
  """Sanity check for API server.

  Args:
    server: str API server to access for this API call.
    version: str API version being used to access the server.

  Raises:
    ValidationError: if the given API server or version is not valid.
  """
  # Map of supported API servers and versions.
  prod = {'v1.19': 'https://advertisersapi.doubleclick.net',
          'v1.20': 'https://advertisersapi.doubleclick.net'}
  beta = {'v1.19': 'https://betaadvertisersapi.doubleclick.net',
          'v1.20': 'https://betaadvertisersapi.doubleclick.net'}
  test = {'v1.19': 'https://advertisersapitest.doubleclick.net',
          'v1.20': 'https://advertisersapitest.doubleclick.net'}

  if (server not in prod.values() and server not in beta.values() and
      server not in test.values()):
    msg = ('Given API server, \'%s\', is not valid. Expecting one of %s.'
           % (server,
              sorted(prod.values() + beta.values() + test.values())[1:]))
    raise ValidationError(msg)

  if (version not in prod.keys() and version not in beta.keys() and
      version not in test.keys()):
    msg = ('Given API version, \'%s\', is not valid. Expecting one of %s.'
           % (version, sorted(set(prod.keys() + beta.keys() + test.keys()))))
    raise ValidationError(msg)

  if (server != prod[version] and server != beta[version] and
      server != test[version]):
    msg = ('Given API version, \'%s\', is not compatible with given server, '
           '\'%s\'.' % (version, server))
    raise ValidationError(msg)



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for accessing all other services."""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

import os
import threading

from adspygoogle.common import SanityCheck
from adspygoogle.common import Utils
from adspygoogle.common.Client import Client
from adspygoogle.common.Logger import Logger
from adspygoogle.dfa import DEFAULT_API_VERSION
from adspygoogle.dfa import DfaSanityCheck
from adspygoogle.dfa import DfaUtils
from adspygoogle.dfa import LIB_SIG
from adspygoogle.dfa import REQUIRED_SOAP_HEADERS
from adspygoogle.dfa.GenericDfaService import GenericDfaService


class DfaClient(Client):

  """Provides entry point to all web services.

  Allows instantiation of all DFA API web services.
  """

  auth_pkl_name = 'dfa_api_auth.pkl'
  config_pkl_name = 'dfa_api_config.pkl'

  def __init__(self, headers=None, config=None, path=None):
    """Inits DfaClient.

    Args:
      [optional]
      headers: dict Object with populated authentication credentials.
      config: dict Object with client configuration values.
      path: str Relative or absolute path to home directory (i.e. location of
            pickles and logs/).

    Example:
      headers = {
        'Username': 'johndoe@example.com',
        'Password': 'secret',
        'AuthToken': '...'
      }
      config = {
        'home': '/path/to/home',
        'log_home': '/path/to/logs/home',
        'xml_parser': '1', # PYXML = 1, ELEMENTREE = 2
        'debug': 'n',
        'raw_debug': 'n',
        'xml_log': 'y',
        'request_log': 'y',
        'raw_response': 'n',
        'strict': 'y',
        'pretty_xml': 'y',
        'compress': 'y',
      }
      path = '/path/to/home'
    """
    super(DfaClient, self).__init__(headers, config, path)

    self.__lock = threading.RLock()
    self.__loc = None

    if path is not None:
      # Update absolute path for a given instance of DfaClient, based on
      # provided relative path.
      if os.path.isabs(path):
        DfaClient.home = path
      else:
        # NOTE(api.sgrinberg): Keep first parameter of join() as os.getcwd(),
        # do not change it to DfaClient.home. Otherwise, may break when
        # multiple instances of DfaClient exist during program run.
        DfaClient.home = os.path.join(os.getcwd(), path)

      # If pickles don't exist at given location, default to "~".
      if (not headers and not config and
          (not os.path.exists(os.path.join(DfaClient.home,
                                           DfaClient.auth_pkl_name)) or
           not os.path.exists(os.path.join(DfaClient.home,
                                           DfaClient.config_pkl_name)))):
        DfaClient.home = os.path.expanduser('~')
    else:
      DfaClient.home = os.path.expanduser('~')

    # Update location for both pickles.
    DfaClient.auth_pkl = os.path.join(DfaClient.home,
                                      DfaClient.auth_pkl_name)
    DfaClient.config_pkl = os.path.join(DfaClient.home,
                                        DfaClient.config_pkl_name)

    # Only load from the pickle if config wasn't specified.
    self._config = config or self.__LoadConfigValues()
    self._config = self.__SetMissingDefaultConfigValues(self._config)
    self._config['home'] = DfaClient.home

    # Validate XML parser to use.
    SanityCheck.ValidateConfigXmlParser(self._config['xml_parser'])

    # Only load from the pickle if 'headers' wasn't specified.
    if headers is None:
      self._headers = self.__LoadAuthCredentials()
    else:
      # Pass app_name from config as appName to headers if still present
      if self._config.get('app_name', None):
        headers['appName'] = self._config['app_name']

      if Utils.BoolTypeConvert(self._config['strict']):
        SanityCheck.ValidateRequiredHeaders(headers, REQUIRED_SOAP_HEADERS)
      self._headers = headers

    # Initialize logger.
    self.__logger = Logger(LIB_SIG, self._config['log_home'])

  def __LoadAuthCredentials(self):
    """Load existing authentication credentials from dfa_api_auth.pkl.

    Returns:
      dict Dictionary object with populated authentication credentials.
    """
    return super(DfaClient, self)._LoadAuthCredentials()

  def __WriteUpdatedAuthValue(self, key, new_value):
    """Write updated authentication value for a key in dfa_api_auth.pkl.

    Args:
      key: str Key to update.
      new_value: str New value to update the key with.
    """
    super(DfaClient, self)._WriteUpdatedAuthValue(key, new_value)

  def __LoadConfigValues(self):
    """Load existing configuration values from dfa_api_config.pkl.

    Returns:
      dict Dictionary object with populated configuration values.
    """
    return super(DfaClient, self)._LoadConfigValues()

  def __SetMissingDefaultConfigValues(self, config=None):
    """Set default configuration values for missing elements in the config dict.

    Args:
      config: dict Object with client configuration values.

    Returns:
      dictionary Configuration values with defaults added in.
    """
    if config is None:
      config = {}
    config = super(DfaClient, self)._SetMissingDefaultConfigValues(config)

    default_config = {
        'home': DfaClient.home,
        'log_home': os.path.join(DfaClient.home, 'logs')
    }
    for key in default_config:
      if key not in config:
        config[key] = default_config[key]
    return config

  def CallRawMethod(self, soap_message, url, server, http_proxy):
    """Call API method directly, using raw SOAP message.

    For API calls performed with this method, outgoing data is not run through
    library's validation logic.

    Args:
      soap_message: str SOAP XML message.
      url: str URL of the API service for the method to call.
      server: str API server to access for this API call.
      http_proxy: str HTTP proxy to use for this API call.

    Returns:
      tuple Response from the API method (SOAP XML response message).
    """
    service_name = DfaUtils.DetermineServiceFromUrl(url).capitalize()
    service = getattr(self, 'Get' + service_name + 'Service')(
        server=server, http_proxy=http_proxy)

    return service.CallRawMethod(soap_message)

  def GetService(self, service_name,
                 server='https://advertisersapi.doubleclick.net', version=None,
                 http_proxy=None, op_config=None):
    """Generic method to create a service.

    Args:
      service_name: str Name of the service to create.
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.
      op_config: dict Dictionary object with additional configuration values for
                 this operation.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      DfaSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    if op_config is None:
      op_config = {
          'server': server,
          'version': version,
          'http_proxy': http_proxy
      }
    return GenericDfaService(self._headers, self._config, op_config,
                             self.__lock, self.__logger, service_name)

  def GetAdService(self, server='https://advertisersapi.doubleclick.net',
                   version=None, http_proxy=None):
    """Returns an object which can call methods in the ad service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('ad', server, version, http_proxy)

  def GetAdvertiserService(self,
                           server='https://advertisersapi.doubleclick.net',
                           version=None, http_proxy=None):
    """Returns an object which can call methods in the advertiser service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('advertiser', server, version, http_proxy)

  def GetAdvertiserGroupService(self,
                                server='https://advertisersapi.doubleclick.net',
                                version=None, http_proxy=None):
    """Returns an object which can call methods in the advertisergroup service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('advertisergroup', server, version, http_proxy)

  def GetCampaignService(self, server='https://advertisersapi.doubleclick.net',
                         version=None, http_proxy=None):
    """Returns an object which can call methods in the campaign service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('campaign', server, version, http_proxy)

  def GetChangeLogService(self, server='https://advertisersapi.doubleclick.net',
                          version=None, http_proxy=None):
    """Returns an object which can call methods in the changelog service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('changelog', server, version, http_proxy)

  def GetContentCategoryService(self,
                                server='https://advertisersapi.doubleclick.net',
                                version=None, http_proxy=None):
    """Returns an object which can call methods in the contentcategory service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('contentcategory', server, version, http_proxy)

  def GetCreativeService(self, server='https://advertisersapi.doubleclick.net',
                         version=None, http_proxy=None):
    """Returns an object which can call methods in the creative service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('creative', server, version, http_proxy)

  def GetCreativeFieldService(self,
                              server='https://advertisersapi.doubleclick.net',
                              version=None, http_proxy=None):
    """Returns an object which can call methods in the creativefield service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('creativefield', server, version, http_proxy)

  def GetCreativeGroupService(self,
                              server='https://advertisersapi.doubleclick.net',
                              version=None, http_proxy=None):
    """Returns an object which can call methods in the creativegroup service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('creativegroup', server, version, http_proxy)

  def GetLoginService(self, server='https://advertisersapi.doubleclick.net',
                      version=None, http_proxy=None):
    """Returns an object which can call methods in the login service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('login', server, version, http_proxy)

  def GetNetworkService(self, server='https://advertisersapi.doubleclick.net',
                        version=None, http_proxy=None):
    """Returns an object which can call methods in the network service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('network', server, version, http_proxy)

  def GetPlacementService(self, server='https://advertisersapi.doubleclick.net',
                          version=None, http_proxy=None):
    """Returns an object which can call methods in the placement service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('placement', server, version, http_proxy)

  def GetReportService(self, server='https://advertisersapi.doubleclick.net',
                       version=None, http_proxy=None):
    """Returns an object which can call methods in the report service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('report', server, version, http_proxy)

  def GetSiteService(self, server='https://advertisersapi.doubleclick.net',
                     version=None, http_proxy=None):
    """Returns an object which can call methods in the site service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('site', server, version, http_proxy)

  def GetSizeService(self, server='https://advertisersapi.doubleclick.net',
                     version=None, http_proxy=None):
    """Returns an object which can call methods in the size service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('size', server, version, http_proxy)

  def GetSpotlightService(self, server='https://advertisersapi.doubleclick.net',
                          version=None, http_proxy=None):
    """Returns an object which can call methods in the spotlight service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('spotlight', server, version, http_proxy)

  def GetStrategyService(self, server='https://advertisersapi.doubleclick.net',
                         version=None, http_proxy=None):
    """Returns an object which can call methods in the strategy service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('strategy', server, version, http_proxy)

  def GetSubnetworkService(self,
                           server='https://advertisersapi.doubleclick.net',
                           version=None, http_proxy=None):
    """Returns an object which can call methods in the subnetwork service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('subnetwork', server, version, http_proxy)

  def GetUserService(self, server='https://advertisersapi.doubleclick.net',
                     version=None, http_proxy=None):
    """Returns an object which can call methods in the user service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('user', server, version, http_proxy)

  def GetUserRoleService(self, server='https://advertisersapi.doubleclick.net',
                         version=None, http_proxy=None):
    """Returns an object which can call methods in the userrole service.

    Args:
      [optional]
      server: str API server this object will access. Possible values are:
              'https://advertisersapi.doubleclick.net' for production,
              'https://advertisersapitest.doubleclick.net' for test, and
              'https://betaadvertisersapi.doubleclick.net' for beta.
              The default behavior is to access the production environment.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfaService New object representing the SOAP service.
    """
    return self.GetService('userrole', server, version, http_proxy)



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for working with reports."""

__author__ = 'api.kwinter@gmail.com (Kevin Winter)'

import datetime
import gzip
import re
import StringIO
import time
import urllib
import urllib2

from adspygoogle import SOAPpy
from adspygoogle.adwords import AUTH_TOKEN_EXPIRE
from adspygoogle.adwords import AUTH_TOKEN_SERVICE
from adspygoogle.adwords import LIB_SIG
from adspygoogle.adwords.AdWordsErrors import AdWordsError
from adspygoogle.adwords.AdWordsErrors import AdWordsReportError
from adspygoogle.adwords.util import XsdToWsdl
from adspygoogle.common import MessageHandler
from adspygoogle.common import SanityCheck
from adspygoogle.common import Utils
from adspygoogle.common.Errors import ValidationError
from adspygoogle.common.Logger import Logger


SERVICE_NAME = 'ReportDefinitionService'
DOWNLOAD_URL_BASE = '/api/adwords/reportdownload'
REPORT_ID='?__rd=%s'
VERSIONED='/%s'
ATTRIBUTES_REGEX = r'( )?[\w:-]+="[\w:\[\]-]+"'
ERROR_TYPE_REGEX = r'(?s)<type>(.*?)</type>'
ERROR_TRIGGER_REGEX = r'(?s)<trigger>(.*?)</trigger>'
ERROR_FIELD_PATH_REGEX = r'(?s)<fieldPath>(.*?)</fieldPath>'
BUF_SIZE = 4096
# We will refresh an OAuth 2.0 credential _OAUTH2_REFRESH_MINUTES_IN_ADVANCE
# minutes in advance of it's expiration.
_OAUTH2_REFRESH_MINUTES_IN_ADVANCE = 5


class ReportDownloader(object):

  """Utility class that downloads reports."""

  def __init__(self, headers, config, op_config, logger):
    """Inits ReportDownloader.

    Args:
      headers: dict Dictionary object with populated authentication
               credentials.
      config: dict Dictionary object with populated configuration values.
      op_config: dict Dictionary object with additional configuration values for
                 this operation.
    """
    self._headers = headers
    self._config = config
    self._op_config = op_config
    self._message_handler = MessageHandler
    namespace_suffix = '/'.join(('/api/adwords', op_config['group'],
                                 self._op_config['version']))
    self._namespace = 'https://adwords.google.com' + namespace_suffix
    xsd_url = '%s%s%s/reportDefinition.xsd' % (op_config['server'],
                                               '/api/adwords/reportdownload/',
                                               self._op_config['version'])
    self._soappyservice = XsdToWsdl.CreateWsdlFromXsdUrl(xsd_url)
    self._logger = logger

  def DownloadReport(self, report_definition_or_id, return_micros=False,
                     file_path=None, fileobj=None):
    """Downloads a report by object or id.

    Args:
      report_definition_or_id: dict or str Report or reportDefinitionId.
      return_micros: bool Whether to return currency in micros (optional).
      file_path: str File path to download to (optional).
      fileobj: file An already-open file-like object that supports write()
               (optional).

    Returns:
      str Report data if file_path and fileobj are None, None if fileobj is
          not None and file_path otherwise.
    """
    if not fileobj and file_path:
      fileobj = open(file_path, 'w+')

    if isinstance(report_definition_or_id, dict):
      return self.__DownloadAdHocReport(report_definition_or_id, return_micros,
                                        fileobj) or file_path
    else:
      return self.__DownloadReportById(report_definition_or_id, return_micros,
                                       fileobj) or file_path

  def DownloadReportWithAwql(self, report_query, download_format,
                             return_micros=False, file_path=None, fileobj=None):
    """Downloads a report with AWQL.

    Args:
      report_query: str AWQL for the report.
      download_format: str Download format. E.g. CSV, TSV, XML.
      return_micros: bool Whether to return currency in micros (optional).
      file_path: str File path to download to (optional).
      fileobj: file An already-open file-like object that supports write()
               (optional).

    Returns:
      str Report data if file_path and fileobj are None, None if fileobj is
          not None and file_path otherwise.
    """
    if not fileobj and file_path:
      fileobj = open(file_path, 'w+')

    return self.__DownloadAdHocReportWithAwql(report_query,
                                              download_format,
                                              return_micros,
                                              fileobj) or file_path

  def __DownloadAdHocReport(self, report_definition, return_micros=False,
                            fileobj=None):
    """Downloads an AdHoc report.

    Args:
      report_definition: dict Report to download.
      return_micros: bool Whether to return currency in micros (optional).
      fileobj: file File to write to (optional).

    Returns:
      str Report data if no fileobj, otherwise None.
    """
    report_xml = self.__GetReportXml(report_definition)
    query_params = {'__rdxml': report_xml}

    payload = urllib.urlencode(query_params)
    return self.__DownloadReport(payload, return_micros, fileobj)

  def __DownloadAdHocReportWithAwql(self,
                                    report_query,
                                    download_format,
                                    return_micros=False,
                                    fileobj=None):
    """Downloads an AdHoc report with AWQL.

    Args:
      report_query: str AWQL to download a report for.
      download_format: str Format of the report download.
      return_micros: bool Whether to return currency in micros (optional).
      fileobj: file File to write to (optional).

    Returns:
      str Report data if no fileobj, otherwise None.
    """
    query_params = {
        '__fmt': download_format,
        '__rdquery': report_query
    }

    payload = urllib.urlencode(query_params)
    return self.__DownloadReport(payload, return_micros, fileobj)

  def __DownloadReport(self, report_payload, return_micros=False, fileobj=None):
    """Downloads an AdHoc report for the specified payload.

    Args:
      report_payload: str Report payload to POST to the server.
      return_micros: bool Whether to return currency in micros (optional).
      fileobj: file File to write to (optional).

    Returns:
      str Report data if no fileobj, otherwise None.
    """
    url = self.__GenerateUrl()
    self._CheckAuthentication()
    headers = self.__GenerateHeaders(return_micros)
    headers['Content-Type'] = 'application/x-www-form-urlencoded'
    headers['Content-Length'] = str(len(report_payload))
    return self.__MakeRequest(url, headers, fileobj, payload=report_payload)

  def __GetReportXml(self, report):
    """Transforms the report object into xml.

    Args:
      report: dict ReportDefinition object to turn to xml.

    Returns:
      str ReportDefinition XML.
    """
    SanityCheck.SoappySanityCheck(self._soappyservice, report, self._namespace,
                                  u'reportDefinition')

    packed = self._message_handler.PackForSoappy(report, self._namespace,
                                                 'reportDefinition',
                                                 self._soappyservice, False,
                                                 lambda x: '')

    # Use a SOAPBuilder
    builder = SOAPpy.SOAPBuilder(kw={'reportDefinition': packed}, envelope=0,
                                 noroot=1)

    # Fixes list serialization.
    builder.config.typed = False

    # Hack, need to remove top element and body wrapper.
    builder._xml_top = ''
    builder.body = 0

    # Build the XML.
    report_xml = builder.build()

    # Removes xsi:types.
    report_xml = self.__RemoveAttributes(report_xml)
    return report_xml

  def __RemoveAttributes(self, report_xml):
    """Removes all attributes from tags.

    Args:
      report_xml: str xml to remove attributes from.

    Returns:
      str Report xml with attributes removed.
    """
    return re.sub(ATTRIBUTES_REGEX, '', report_xml).strip()

  def __DownloadReportById(self, report_definition_id, return_micros=False,
                           fileobj=None):
    """Download report and return raw data.

    Args:
      report_definition_id: str Id of the report definition to download.
      return_micros: bool Whether to return currency in micros.
      fileobj: str Path to download file to.

    Returns:
      str Report data if no fileobj, otherwise None.
    """
    self._CheckAuthentication()
    url = self.__GenerateUrl(report_definition_id)
    headers = self.__GenerateHeaders(return_micros)
    return self.__MakeRequest(url, headers, fileobj)

  def __GenerateUrl(self, report_definition_id=None):
    """Generates the URL to get a report from.

    Args:
      report_definition_id: int ID of the report to download.

    Returns:
      str url to request
    """
    url = [DOWNLOAD_URL_BASE]
    url.append(VERSIONED % self._op_config['version'])
    if report_definition_id:
      url.append(REPORT_ID % report_definition_id)
    return ''.join(url)

  def __GenerateHeaders(self, return_micros):
    """Generates the headers to use for the report download.

    Args:
      return_micros: bool whether or not to use micros for money.

    Returns:
      dict Dictionary containing all the headers for the request
    """
    headers = {}
    if 'clientCustomerId' in self._headers:
      headers['clientCustomerId'] = self._headers['clientCustomerId']

    # Handle OAuth (if enabled) and ClientLogin
    if self._headers.get('oauth2credentials'):
      self._headers['oauth2credentials'].apply(headers)
    else:
      headers['Authorization'] = ('GoogleLogin %s' %
          urllib.urlencode({'auth':
                            self._headers['authToken'].strip()}))

    headers['returnMoneyInMicros'] = str(return_micros).lower()
    headers['developerToken'] = self._headers['developerToken']
    headers['User-Agent'] = self._headers['userAgent']
    if Utils.BoolTypeConvert(self._config['compress']):
      headers['Accept-Encoding'] = 'gzip'
      headers['User-Agent'] += ',gzip'
      headers['Content-Encoding'] = 'gzip'
    return headers

  def __MakeRequest(self, url, headers=None, fileobj=None, payload=None):
    """Performs an HTTPS request and slightly processes the response.

    If fileobj is provided, saves the body to file instead of including it
    in the return value.

    Args:
      url: str Resource for the request line.
      headers: dict Headers to send along with the request.
      fileobj: file File to save to (optional).
      payload: str Xml to POST (optional).

    Returns:
      str Report data as a string if fileobj=None, otherwise None
    """
    headers = headers or {}
    request_url = self._op_config['server'] + url

    orig_payload = payload

    if Utils.BoolTypeConvert(self._config['compress']):
      buffer = StringIO.StringIO()
      gzip_file = gzip.GzipFile(mode='wb', fileobj=buffer)
      gzip_file.write(payload)
      gzip_file.close()
      payload = buffer.getvalue()
      headers['Content-Length'] = str(len(payload))

    start_time = time.strftime('%Y-%m-%d %H:%M:%S')
    request = urllib2.Request(request_url, payload, headers)
    try:
      response_code = '---'
      response_headers = []
      try:
        response = urllib2.urlopen(request)
        response_code = response.code
        response_headers = response.info().headers
        if response.info().get('Content-Encoding') == 'gzip':
          response = gzip.GzipFile(fileobj=StringIO.StringIO(response.read()),
                                   mode='rb')
        if fileobj:
          self.__DumpToFile(response, fileobj)
          return None
        else:
          return response.read()
      except urllib2.HTTPError, e:
        response = e
        response_code = response.code
        response_headers = response.info().headers
        if response.info().get('Content-Encoding') == 'gzip':
          response = gzip.GzipFile(fileobj=StringIO.StringIO(response.read()),
                                   mode='rb')
        error = response.read()
        self.__CheckForXmlError(response_code, error)
        raise AdWordsError('%s %s' % (str(e), error))
      except urllib2.URLError, e:
        raise AdWordsError(str(e))
    finally:
      end_time = time.strftime('%Y-%m-%d %H:%M:%S')
      xml_log_data = self.__CreateXmlLogData(start_time, end_time, request_url,
                                             headers, orig_payload,
                                             response_code, response_headers)
      self.__LogRequest(xml_log_data)

  def __CheckForXmlError(self, response_code, response):
    if 'reportDownloadError' in response:
      error_type = re.search(ERROR_TYPE_REGEX, response)
      if error_type: error_type = error_type.group(1)
      trigger = re.search(ERROR_TRIGGER_REGEX, response)
      if trigger: trigger = trigger.group(1)
      field_path = re.search(ERROR_FIELD_PATH_REGEX, response)
      if field_path: field_path = field_path.group(1)
      raise AdWordsReportError(response_code, error_type, trigger, field_path)

  def _CheckAuthentication(self):
    """Ensures we have authentication values ready to make the request."""
    if self._headers.get('oauth2credentials'):
      self._RefreshCredentialIfNecessary(self._headers['oauth2credentials'])
    else:
      self.__ReloadAuthToken()

  def _RefreshCredentialIfNecessary(self, credential):
    """Checks if the credential needs refreshing and refreshes if necessary."""
    if (credential.token_expiry is not None and credential.token_expiry -
        datetime.datetime.utcnow() <
        datetime.timedelta(minutes=_OAUTH2_REFRESH_MINUTES_IN_ADVANCE)):
      import httplib2
      self._headers['oauth2credentials'].refresh(httplib2.Http())

  def __ReloadAuthToken(self):
    """Ensures we have a valid auth_token in our headers."""
    # Load/set authentication token. If authentication token has expired,
    # regenerate it.
    now = time.time()
    if (('authToken' not in self._headers and
         'auth_token_epoch' not in self._config) or
        int(now - self._config['auth_token_epoch']) >= AUTH_TOKEN_EXPIRE):
      if ('email' not in self._headers or
          not self._headers['email'] or
          'password' not in self._headers or
          not self._headers['password']):
        msg = ('Required authentication headers, \'email\' and \'password\', '
               'are missing. Unable to regenerate authentication token.')
        raise ValidationError(msg)
      self._headers['authToken'] = Utils.GetAuthToken(
          self._headers['email'], self._headers['password'],
          AUTH_TOKEN_SERVICE, LIB_SIG, self._config['proxy'])
      self._config['auth_token_epoch'] = time.time()

  def __DumpToFile(self, response, fileobj):
    """Reads from response.read() and writes to fileobj.

     Args:
      response: file Some object that supports read().
      fileobj: file Some object that supports write()

     Returns:
      number Number of bytes written.
    """
    byteswritten = 0
    while True:
      buf = response.read(BUF_SIZE)
      if buf:
        fileobj.write(buf)
        byteswritten += len(buf)
      else:
        break
    return byteswritten

  def __LogRequest(self, xml_log_data):
    """Logs the Report Download request.

    Args:
      xml_log_data: str Data to log for this request.
    """
    log_handlers = self.__GetLogHandlers()
    for handler in log_handlers:
      handler['target'] = Logger.NONE
      if handler['tag'] == 'xml_log':
        handler['data'] += xml_log_data
    for handler in log_handlers:
      if (handler['tag'] and
          Utils.BoolTypeConvert(self._config[handler['tag']])):
        handler['target'] = Logger.FILE
      # If debugging is On, raise handler's target two levels,
      #   NONE -> CONSOLE
      #   FILE -> FILE_AND_CONSOLE.
      if Utils.BoolTypeConvert(self._config['debug']):
        handler['target'] += 2

      if (handler['target'] != Logger.NONE and handler['data'] and
          handler['data'] != 'None' and handler['data'] != 'DEBUG: '):
        self._logger.Log(handler['name'], handler['data'],
                         log_level=Logger.DEBUG, log_handler=handler['target'])

  def __GetLogHandlers(self):
    """Gets a list of log handlers for the AdWords library.

    Returns:
      list Log handlers for the AdWords library.
    """
    return [
        {
            'tag': 'xml_log',
            'name': 'soap_xml',
            'data': ''
        },
        {
            'tag': 'request_log',
            'name': 'request_info',
            'data': ('host=%s operation=%s'
                     % (self._op_config['server'], 'ReportDownload'))
        }
    ]

  def __CreateXmlLogData(self, start_time, end_time, request_url,
                         request_headers, payload, response_code,
                         response_headers):
    """Transforms arguments into a string to log.

    Args:
      start_time: str Formatted start time.
      end_time: str Formatted end time.
      request_url: str URL POSTed to.
      request_headers: dict Request headers sent with request.
      payload: str Payload (http request body).
      response_code: int Response code from remote server.
      response_headers: list List of string headers received.

    Returns:
      str Data to log.
    """
    # Errors will have a response code, otherwise it should have been success.
    response_code = response_code or 200
    return ('StartTime: %s\n%s\n%s\n\n%s\n\n%s\n%s\nEndTime: %s'
            % (start_time, 'POST: %s' % request_url,
               self.__SerializeRequestHeaders(request_headers), payload,
               'HTTP %s' % response_code,
               self.__SerializeResponseHeaders(response_headers), end_time))

  def __SerializeRequestHeaders(self, request_headers):
    """Serializes the request headers into a string for logging.

    Returns each key->value pair as "key: value" with newlines separating them.

    Args:
      request_headers: dict Dictionary of headers to serialize.

    Returns:
      str Serialized request headers.
    """
    return '\n'.join(['%s: %s' % (key, request_headers[key])
                      for key in request_headers])

  def __SerializeResponseHeaders(self, response_headers):
    """Serializes the response headers.

    Headers are already formatted, this joins them into a single string.

    Args:
      response_headers: list List of string headers received.

    Returns:
      str Serialized response headers.
    """
    return (''.join(response_headers)).strip()



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for accessing all other services."""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

import os
import thread
import time

from adspygoogle.common import SanityCheck
from adspygoogle.common import Utils
from adspygoogle.common.Client import Client
from adspygoogle.common.Errors import AuthTokenError
from adspygoogle.common.Errors import ValidationError
from adspygoogle.common.Logger import Logger
from adspygoogle.dfp import AUTH_TOKEN_SERVICE
from adspygoogle.dfp import DEFAULT_API_VERSION
from adspygoogle.dfp import DfpSanityCheck
from adspygoogle.dfp import LIB_SIG
from adspygoogle.dfp import REQUIRED_SOAP_HEADERS
from adspygoogle.dfp.GenericDfpService import GenericDfpService


class DfpClient(Client):
  """Provides entry point to all web services.

  Allows instantiation of all DFP API web services.
  """

  auth_pkl_name = 'dfp_api_auth.pkl'
  config_pkl_name = 'dfp_api_config.pkl'

  def __init__(self, headers=None, config=None, path=None):
    """Inits Client.

    Args:
      [optional]
      headers: dict Object with populated authentication credentials.
      config: dict Object with client configuration values.
      path: str Relative or absolute path to home directory (i.e. location of
            pickles and logs/).

    Example:
      headers = {
        'email': 'johndoe@example.com',
        'password': 'secret',
        'authToken': '...',
        'applicationName': 'GoogleTest',
        'networkCode': 'ca-01234567',
        'oauth2credentials': 'See use_oauth2.py'
      }
      config = {
        'home': '/path/to/home',
        'log_home': '/path/to/logs/home',
        'proxy': 'http://example.com:8080',
        'xml_parser': '1', # PYXML = 1, ELEMENTREE = 2
        'debug': 'n',
        'raw_debug': 'n',
        'xml_log': 'y',
        'request_log': 'y',
        'raw_response': 'n',
        'strict': 'y',
        'pretty_xml': 'y',
        'compress': 'y',
        'access': ''
      }
      path = '/path/to/home'
    """
    super(DfpClient, self).__init__(headers, config, path)

    self.__lock = thread.allocate_lock()
    self.__loc = None

    if path is not None:
      # Update absolute path for a given instance of DfpClient, based on
      # provided relative path.
      if os.path.isabs(path):
        DfpClient.home = path
      else:
        # NOTE(api.sgrinberg): Keep first parameter of join() as os.getcwd(),
        # do not change it to DfpClient.home. Otherwise, may break when
        # multiple instances of DfpClient exist during program run.
        DfpClient.home = os.path.join(os.getcwd(), path)

      # If pickles don't exist at given location, default to "~".
      if (not headers and not config and
          (not os.path.exists(os.path.join(DfpClient.home,
                                           DfpClient.auth_pkl_name)) or
           not os.path.exists(os.path.join(DfpClient.home,
                                           DfpClient.config_pkl_name)))):
        DfpClient.home = os.path.expanduser('~')
    else:
      DfpClient.home = os.path.expanduser('~')

    # Update location for both pickles.
    DfpClient.auth_pkl = os.path.join(DfpClient.home,
                                      DfpClient.auth_pkl_name)
    DfpClient.config_pkl = os.path.join(DfpClient.home,
                                        DfpClient.config_pkl_name)

    # Only load from the pickle if config wasn't specified.
    self._config = config or self.__LoadConfigValues()
    self._config = self.__SetMissingDefaultConfigValues(self._config)
    self._config['home'] = DfpClient.home

    # Validate XML parser to use.
    SanityCheck.ValidateConfigXmlParser(self._config['xml_parser'])

    # Only load from the pickle if 'headers' wasn't specified.
    if headers is None:
      self._headers = self.__LoadAuthCredentials()
    else:
      if Utils.BoolTypeConvert(self._config['strict']):
        SanityCheck.ValidateRequiredHeaders(headers, REQUIRED_SOAP_HEADERS)
      self._headers = headers

    # Load/set authentication token.
    try:
      if headers and 'authToken' in headers and headers['authToken']:
        self._headers['authToken'] = headers['authToken']
      elif 'email' in self._headers and 'password' in self._headers:
        self._headers['authToken'] = Utils.GetAuthToken(
            self._headers['email'], self._headers['password'],
            AUTH_TOKEN_SERVICE, LIB_SIG, self._config['proxy'])
      elif (self._headers.get('oauth2credentials')):
        # If they have oauth2credentials, that's also fine.
        pass
      else:
        msg = ('Authentication data, email or/and password, OAuth2 credentials '
               'is missing.')
        raise ValidationError(msg)
      self._config['auth_token_epoch'] = time.time()
    except AuthTokenError:
      # We would end up here if non-valid Google Account's credentials were
      # specified.
      self._headers['authToken'] = None
      self._config['auth_token_epoch'] = 0

    # Initialize logger.
    self.__logger = Logger(LIB_SIG, self._config['log_home'])

  def __LoadAuthCredentials(self):
    """Load existing authentication credentials from dfp_api_auth.pkl.

    Returns:
      dict Dictionary object with populated authentication credentials.
    """
    return super(DfpClient, self)._LoadAuthCredentials()

  def __WriteUpdatedAuthValue(self, key, new_value):
    """Write updated authentication value for a key in dfp_api_auth.pkl.

    Args:
      key: str Key to update.
      new_value: str New value to update the key with.
    """
    super(DfpClient, self)._WriteUpdatedAuthValue(key, new_value)

  def __LoadConfigValues(self):
    """Load existing configuration values from dfp_api_config.pkl.

    Returns:
      dict Dictionary object with populated configuration values.
    """
    return super(DfpClient, self)._LoadConfigValues()

  def __SetMissingDefaultConfigValues(self, config=None):
    """Set default configuration values for missing elements in the config dict.

    Args:
      config: dict Object with client configuration values.

    Returns:
      dict The config dictionary with default values added.
    """
    if config is None: config = {}
    config = super(DfpClient, self)._SetMissingDefaultConfigValues(config)
    default_config = {
        'home': DfpClient.home,
        'log_home': os.path.join(DfpClient.home, 'logs')
    }
    for key in default_config:
      if key not in config:
        config[key] = default_config[key]
    return config


  def CallRawMethod(self, soap_message, url, server, http_proxy):
    """Call API method directly, using raw SOAP message.

    For API calls performed with this method, outgoing data is not run through
    library's validation logic.

    Args:
      soap_message: str SOAP XML message.
      url: str URL of the API service for the method to call.
      server: str API server to access for this API call.
      http_proxy: str HTTP proxy to use for this API call.

    Returns:
      tuple Response from the API method (SOAP XML response message).
    """
    service_name = url.split('/')[-1]
    service = getattr(self, 'Get' + service_name)(server=server,
                                                  http_proxy=http_proxy)
    return service.CallRawMethod(soap_message)

  def GetService(self, service_name, server='https://www.google.com',
                 version=None, http_proxy=None, op_config=None):
    """Generic method to create a service.

    Args:
      service_name: str Name of the service to create.
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.
      op_config: dict service configuration.

    Returns:
      GenericDfpService New object representing the SOAP service.
    """
    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      DfpSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    if op_config is None:
      op_config = {
          'server': server,
          'version': version,
          'http_proxy': http_proxy
      }
    return GenericDfpService(self._headers, self._config, op_config,
                             self.__lock, self.__logger, service_name)

  def GetCompanyService(self, server='https://www.google.com', version=None,
                        http_proxy=None):
    """Create a CompanyService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the CompanyService object.
    """
    return self.GetService('CompanyService', server, version, http_proxy)

  def GetContentService(self, server='https://www.google.com', version=None,
                        http_proxy=None):
    """Create a ContentService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the ContentService object.
    """
    return self.GetService('ContentService', server, version, http_proxy)

  def GetCreativeService(self, server='https://www.google.com',
                         version=None, http_proxy=None):
    """Create a CreativeService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the CreativeService
      object.
    """
    return self.GetService('CreativeService', server, version, http_proxy)

  def GetCreativeTemplateService(self, server='https://www.google.com',
                                 version=None, http_proxy=None):
    """Create a CreativeTemplateService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the CreativeTemplateService
      object.
    """
    return self.GetService('CreativeTemplateService', server, version,
                           http_proxy)

  def GetCustomTargetingService(self, server='https://www.google.com',
                                version=None, http_proxy=None):
    """Create a CustomTargetingService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the CustomTargetingService
      object.
    """
    return self.GetService('CustomTargetingService', server, version,
                           http_proxy)

  def GetForecastService(self, server='https://www.google.com',
                         version=None, http_proxy=None):
    """Create a ForecastService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the ForecastService object.
    """
    return self.GetService('ForecastService', server, version, http_proxy)

  def GetInventoryService(self, server='https://www.google.com',
                          version=None, http_proxy=None):
    """Create a InventoryService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the InventoryService object.
    """
    return self.GetService('InventoryService', server, version, http_proxy)

  def GetLabelService(self, server='https://www.google.com',
                      version=None, http_proxy=None):
    """Create a LabelService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the LabelService object.
    """
    return self.GetService('LabelService', server, version, http_proxy)

  def GetLineItemCreativeAssociationService(self,
                                            server='https://www.google.com',
                                            version=None, http_proxy=None):
    """Create a LineItemCreativeAssociationService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the
      LineItemCreativeAssociationService object.
    """
    return self.GetService('LineItemCreativeAssociationService', server,
                           version, http_proxy)

  def GetLineItemService(self, server='https://www.google.com',
                         version=None, http_proxy=None):
    """Create a LineItemService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the LineItemService object.
    """
    return self.GetService('LineItemService', server, version, http_proxy)

  def GetNetworkService(self, server='https://www.google.com', version=None,
                        http_proxy=None):
    """Create a NetworkService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the NetworkService object.
    """
    return self.GetService('NetworkService', server, version, http_proxy)

  def GetOrderService(self, server='https://www.google.com', version=None,
                      http_proxy=None):
    """Create a OrderService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the OrderService object.
    """
    return self.GetService('OrderService', server, version, http_proxy)

  def GetPlacementService(self, server='https://www.google.com',
                          version=None, http_proxy=None):
    """Create a PlacementService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the PlacementService object.
    """
    return self.GetService('PlacementService', server, version, http_proxy)

  def GetPublisherQueryLanguageService(self,
                                       server='https://www.google.com',
                                       version=None, http_proxy=None):
    """Create a PublisherQueryLanguageService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the
      PublisherQueryLanguageService object.
    """
    return self.GetService('PublisherQueryLanguageService', server, version,
                           http_proxy)

  def GetReportService(self, server='https://www.google.com',
                       version=None, http_proxy=None):
    """Create a ReportService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the ReportService object.
    """
    return self.GetService('ReportService', server, version, http_proxy)

  def GetSuggestedAdUnitService(self, server='https://www.google.com',
                                version=None, http_proxy=None):
    """Create a SuggestedAdUnitService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the SuggestedAdUnitService
      object.
    """
    return self.GetService('SuggestedAdUnitService', server, version,
                           http_proxy)

  def GetThirdPartySlotService(self, server='https://www.google.com',
                               version=None, http_proxy=None):
    """Create a ThirdPartySlotService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the ThirdPartySlotService
      object.
    """
    return self.GetService('ThirdPartySlotService', server, version, http_proxy)

  def GetUserService(self, server='https://www.google.com', version=None,
                     http_proxy=None):
    """Create a UserService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the UserService object.
    """
    return self.GetService('UserService', server, version, http_proxy)

  def GetActivityService(self, server='https://www.google.com', version=None,
                         http_proxy=None):
    """Create an ActivityService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the ActivityService object.
    """
    return self.GetService('ActivityService', server, version, http_proxy)

  def GetActivityGroupService(self, server='https://www.google.com',
                              version=None, http_proxy=None):
    """Create an ActivityGroupService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the ActivityGroupService object.
    """
    return self.GetService('ActivityGroupService', server, version, http_proxy)

  def GetAdRuleService(self, server='https://www.google.com', version=None,
                       http_proxy=None):
    """Create an AdRuleService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the AdRuleService object.
    """
    return self.GetService('AdRuleService', server, version, http_proxy)

  def GetContentBundleService(self, server='https://www.google.com',
                              version=None, http_proxy=None):
    """Create an ContentBundleService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the ContentBundleService object.
    """
    return self.GetService('ContentBundleService', server, version, http_proxy)

  def GetContactService(self, server='https://www.google.com',
                        version=None, http_proxy=None):
    """Create an ContactService.

    Args:
      [optional]
      server: str API server to access for API calls. The default value is
              'https://www.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericDfpService New object representing the ContactService object.
    """
    return self.GetService('ContactService', server, version, http_proxy)



#!/usr/bin/python
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation functions."""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

from adspygoogle.common.Errors import ValidationError

DEPRECATED_AFTER = {
    'UserListService': 'v201306',
    'AdExtensionOverrideService': 'v201306'
}

# A map of the supported servers for each version.
_VERSION_SERVER_MAP = {
    'v201306': ('https://adwords.google.com',),
    'v201309': ('https://adwords.google.com',),
}


def ValidateServer(server, version):
  """Sanity check for API server.

  Args:
    server: str API server to access for this API call.
    version: str API version being used to access the server.

  Raises:
    ValidationError: if the given API server or version is not valid.
  """
  if version not in _VERSION_SERVER_MAP:
    raise ValidationError(
        'Given API version, \'%s\', is not valid. Expecting one of %s.'
        % (version, sorted(_VERSION_SERVER_MAP.keys())))

  if server not in _VERSION_SERVER_MAP[version]:
    raise ValidationError(
        'Given API server, \'%s\', is not a valid server for version \'%s\'. '
        'Expecting one of %s.'
        % (server, version, sorted(_VERSION_SERVER_MAP[version])))


def ValidateService(service, version):
  """Checks if this service is available in the requested version.

  Args:
    service: str Service being requested.
    version: str Version being requested.
  """
  if service in DEPRECATED_AFTER and version > DEPRECATED_AFTER[service]:
    raise ValidationError('%s is not available in %s' % (service, version))



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for handling errors."""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

from adspygoogle.common.Errors import Error


class DfaError(Error):

  """Implements a DfaError, responsible for wrapping errors."""

  pass


class DfaApiError(DfaError):

  """Implements a DfaApiError, responsible for handling API exception."""

  def __init__(self, fault):
    (self.__fault, self.fault_code, self.fault_string) = (fault, '', '')
    if 'faultcode' in fault: self.fault_code = fault['faultcode']
    if 'faultstring' in fault: self.fault_string = fault['faultstring']

    (self.code, self.error_message, self.localized_message, self.message,
     self.hostname) = (-1, '', '', '', '')
    if 'detail' in fault and 'doubleclick' in fault['detail']:
      if 'errorCode' in fault['detail']['doubleclick']:
        self.code = int(fault['detail']['doubleclick']['errorCode'])
      if 'errorMessage' in fault['detail']['doubleclick']:
        self.error_message = fault['detail']['doubleclick']['errorMessage']
      if 'localizedMessage' in fault['detail']['doubleclick']:
        self.localized_message = fault['detail']['doubleclick'][
            'localizedMessage']
      if 'message' in fault['detail']['doubleclick']:
        self.message = fault['detail']['doubleclick']['message']
      if 'hostname' in fault['detail']:
        self.hostname = fault['detail']['hostname']
    if not self.message: self.message = self.fault_string

  def __str__(self):
    if self.code > -1:
      return 'Code %s: %s' % (self.code, self.message)
    else:
      return self.fault_string

  def __call__(self):
    return (self.__fault,)


class DfaAuthenticationError(DfaApiError):

  """Implements DfaAuthenticationError.

  Responsible for handling authentication error.
  """

  pass



#!/usr/bin/python
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for generating authentication token to access Google Account."""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

import os
import urllib
import urllib2
import warnings

from adspygoogle.common.Errors import AuthTokenError
from adspygoogle.common.Errors import CaptchaError


_DEPRECATION_WARNING = 'ClientLogin is deprecated. Please use OAuth 2.0'
warnings.filterwarnings('always', _DEPRECATION_WARNING, DeprecationWarning)


class AuthToken(object):

  """Fetches authentication token.

  Responsible for generating the authentication token to access Google Account,
  https://www.google.com/accounts/NewAccount. The token is fetched via the
  ClientLogin API, http://code.google.com/apis/accounts/.
  """

  def __init__(self, email, password, service, lib_sig, proxy, login_token=None,
               login_captcha=None):
    """Inits AuthToken.

    Args:
      email: str Login email of the Google Account.
      password: str Login password of the Google Account.
      service: str Name of the Google service for which to authorize access.
      lib_sig: str Client library signature.
      proxy: str HTTP proxy to use.
    """
    self.__email = email
    self.__password = password
    self.__account_type = 'GOOGLE'
    self.__service = service
    self.__source = 'Google-%s' % lib_sig
    self.__proxy = proxy
    self.__sid = ''
    self.__lsid = ''
    self.__auth = ''
    self.__login_token = login_token
    self.__login_captcha = login_captcha

    self.__Login()

  def __Login(self):
    """Fetch Auth token and SID, LSID cookies from Google Account auth."""
    warnings.warn(_DEPRECATION_WARNING, DeprecationWarning, stacklevel=5)
    if self.__proxy: os.environ['http_proxy'] = self.__proxy
    url = 'https://www.google.com/accounts/ClientLogin'
    data = [('Email', self.__email),
            ('Passwd', self.__password),
            ('accountType', self.__account_type),
            ('service', self.__service),
            ('source', self.__source)]
    if self.__login_token and self.__login_captcha:
      data.append(('logintoken', self.__login_token))
      data.append(('logincaptcha', self.__login_captcha))
    try:
      fh = urllib2.urlopen(url, urllib.urlencode(data))
      data = self.__ParseResponse(fh)
      try:
        if 'SID' in data or 'LSID' in data or 'Auth' in data:
          self.__sid = data.get('SID', '')
          self.__lsid = data.get('LSID', '')
          self.__auth = data.get('Auth', '')
        elif 'Error' in data and data['Error'] != 'CaptchaRequired':
          msg = data['Error'].strip()
          if 'Info' in data:
            msg = msg + ' Additional Info: ' + data['Info'].strip()
          raise AuthTokenError(msg)
        elif 'CaptchaToken' in data:
          raise CaptchaError(data['CaptchaToken'].strip(),
                             'http://www.google.com/accounts/'
                             + data['CaptchaUrl'].strip())
        else:
          raise AuthTokenError('Unexpected response: ' + str(data))
      finally:
        fh.close()
    except IOError, e:
      raise AuthTokenError(e)

  def __ParseResponse(self, fh):
    """Processes the ClientLogin response into a dict.

    Args:
      fh: file The file object representing the ClientLogin response.

    Returns:
      dict Dictionary containing the response in key-value format.
    """
    ret = {}
    for line in fh:
      index = line.index('=')
      key, value = line[:index], line[index + 1:].strip()
      ret[key] = value
    return ret

  def GetSidToken(self):
    """Return SID cookie.

    Returns:
      str SID cookie.
    """
    return self.__sid

  def GetLsidToken(self):
    """Return LSID cookie.

    Returns:
      str LSDI cookie.
    """
    return self.__lsid

  def GetAuthToken(self):
    """Return Auth authentication token.

    Returns:
      str Auth authentication token.
    """
    return self.__auth



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic proxy to access any AdWords web service."""

__author__ = 'api.jdilallo@gmail.com (Joseph DiLallo)'

import time

from adspygoogle import SOAPpy
from adspygoogle.adwords import AdWordsUtils
from adspygoogle.adwords import AUTH_TOKEN_EXPIRE
from adspygoogle.adwords import AUTH_TOKEN_SERVICE
from adspygoogle.adwords import LIB_SIG
from adspygoogle.adwords import LIB_URL
from adspygoogle.adwords.AdWordsErrors import AdWordsApiError
from adspygoogle.adwords.AdWordsErrors import AdWordsError
from adspygoogle.adwords.AdWordsErrors import ERRORS
from adspygoogle.adwords.AdWordsSoapBuffer import AdWordsSoapBuffer
from adspygoogle.common import Utils
from adspygoogle.common.Errors import Error
from adspygoogle.common.Errors import ValidationError
from adspygoogle.common.GenericApiService import GenericApiService
from adspygoogle.common.GenericApiService import MethodInfoKeys


class GenericAdWordsService(GenericApiService):

  """Wrapper for any AdWords web service."""

  # The _POSSIBLE_ADWORDS_REQUEST_HEADERS are both the SOAP element names and
  # the self._headers dictionary keys for all elements that may be in an AdWords
  # header.
  _POSSIBLE_ADWORDS_REQUEST_HEADERS = (
      'authToken', 'developerToken', 'userAgent', 'clientCustomerId',
      'validateOnly', 'partialFailure')
  # The _OAUTH_IGNORE_HEADERS are the header elements that should not be
  # included when the client is using OAuth.
  _OAUTH_IGNORE_HEADERS = ('authToken')
  # The _WRAP_LISTS constant indicates that AdWords services do not need to wrap
  # lists in an extra layer of XML element tags.
  _WRAP_LISTS = False
  # The _BUFFER_CLASS is the subclass of SoapBuffer that should be used to track
  # all SOAP interactions
  _BUFFER_CLASS = AdWordsSoapBuffer
  # List of fields we should convert to string.
  _STR_CONVERT = ['clientCustomerId']

  def __init__(self, headers, config, op_config, lock, logger, service_name):
    """Inits GenericAdWordsService.

    Args:
      headers: dict Dictionary object with populated authentication
               credentials.
      config: dict Dictionary object with populated configuration values.
      op_config: dict Dictionary object with additional configuration values for
                 this operation.
      lock: thread.lock Thread lock to use to synchronize requests.
      logger: Logger Instance of Logger to use for logging.
      service_name: string The name of this service.
    """
    group = op_config['group']
    if service_name == 'BulkMutateJobService': group = 'job'
    service_url = [op_config['server'], 'api/adwords', group,
                   op_config['version'], service_name]
    if config['access']:
      service_url.insert(len(service_url) - 1, config['access'])
    service_url = '/'.join(service_url)
    namespace = '/'.join(['https://adwords.google.com/api/adwords',
                          op_config['group'], op_config['version']])
    namespace_extractor = _DetermineNamespacePrefix

    super(GenericAdWordsService, self).__init__(
        headers, config, op_config, lock, logger, service_name, service_url,
        GenericAdWordsService._WRAP_LISTS, GenericAdWordsService._BUFFER_CLASS,
        namespace, namespace_extractor)

    # AdWords-specific changes to the SOAPpy.WSDL.Proxy
    methodattrs = {}
    for namespace in self._soappyservice.wsdl.types.keys():
      group_name = AdWordsUtils.ExtractGroupNameFromUrl(namespace)
      methodattrs['xmlns:' + group_name] = namespace
    methodattrs['xmlns'] = self._namespace
    self._soappyservice.soapproxy.methodattrs = methodattrs

  def _SetHeaders(self):
    """Sets the SOAP headers for this service's requests."""
    now = time.time()
    if ((('authToken' not in self._headers and
          'auth_token_epoch' not in self._config) or
         int(now - self._config['auth_token_epoch']) >= AUTH_TOKEN_EXPIRE) and
        not self._headers.get('oauth2credentials')):
      if ('email' not in self._headers or not self._headers['email'] or
          'password' not in self._headers or not self._headers['password']):
        raise ValidationError('Required authentication headers, \'email\' and '
                              '\'password\', are missing. Unable to regenerate '
                              'authentication token.')
      self._headers['authToken'] = Utils.GetAuthToken(
          self._headers['email'], self._headers['password'], AUTH_TOKEN_SERVICE,
          LIB_SIG, self._config['proxy'])
      self._config['auth_token_epoch'] = time.time()

    # Apply headers to the SOAPpy service.
    header_attrs = {
        'xmlns': self._namespace,
        'xmlns:cm': ('https://adwords.google.com/api/adwords/cm/' +
                     self._op_config['version'])
    }
    soap_headers = SOAPpy.Types.headerType(attrs=header_attrs)
    request_header_data = {}
    for key in GenericAdWordsService._POSSIBLE_ADWORDS_REQUEST_HEADERS:
      if (key in GenericAdWordsService._OAUTH_IGNORE_HEADERS
          and self._headers.get('oauth2credentials')):
        continue
      if key in self._headers and self._headers[key]:
        value = self._headers[key]
        if key in GenericAdWordsService._STR_CONVERT:
          value = str(value)
        elif key == 'userAgent':
          value = ''.join([value, LIB_SIG])
        request_header_data['cm:' + key] = SOAPpy.Types.stringType(
            value)
    request_header = SOAPpy.Types.structType(
        data=request_header_data, name='RequestHeader', typed=0)
    soap_headers.RequestHeader = request_header
    self._soappyservice.soapproxy.header = soap_headers

  def _GetMethodInfo(self, method_name):
    """Pulls all of the relevant data about a method from a SOAPpy service.

    The return dictionary has two keys, MethodInfoKeys.INPUTS and
    MethodInfoKeys.OUTPUTS. Each of these keys has a list value. These lists
    contain a dictionary of information on the input/output parameter list, in
    order.

    Args:
      method_name: string The name of the method to pull information for.
    Returns:
      dict A dictionary containing information about a SOAP method.
    """
    rval = {}
    rval[MethodInfoKeys.INPUTS] = []
    for i in range(len(self._soappyservice.wsdl.types[
        self._namespace].elements[method_name].content.content.content)):
      param_attributes = self._soappyservice.wsdl.types[
          self._namespace].elements[method_name].content.content.content[
              i].attributes
      inparam = {
          MethodInfoKeys.ELEMENT_NAME: param_attributes['name'],
          MethodInfoKeys.NS: param_attributes['type'].getTargetNamespace(),
          MethodInfoKeys.TYPE: param_attributes['type'].getName(),
          MethodInfoKeys.MAX_OCCURS: param_attributes['maxOccurs']
      }
      rval[MethodInfoKeys.INPUTS].append(inparam)

    rval[MethodInfoKeys.OUTPUTS] = []
    for i in range(len(self._soappyservice.wsdl.types[
        self._namespace].elements[
            method_name + 'Response'].content.content.content)):
      param_attributes = self._soappyservice.wsdl.types[
          self._namespace].elements[
              method_name + 'Response'].content.content.content[i].attributes
      outparam = {
          MethodInfoKeys.ELEMENT_NAME: param_attributes['name'],
          MethodInfoKeys.NS: param_attributes['type'].getTargetNamespace(),
          MethodInfoKeys.TYPE: param_attributes['type'].getName(),
          MethodInfoKeys.MAX_OCCURS: param_attributes['maxOccurs']
      }
      rval[MethodInfoKeys.OUTPUTS].append(outparam)
    return rval

  def _TakeActionOnSoapCall(self, method_name, args):
    """Gives the service a chance to take product-specific action on raw inputs.

    AdWords will support legacy xsi_typing for the BulkMutateJobService.

    Args:
      method_name: string The name of the SOAP operation being called.
      args: tuple The arguments passed into the SOAP operation.

    Returns:
      tuple The method arguments, possibly modified.
    """
    if (self._service_name == 'BulkMutateJobService' and
        method_name.lower() == 'mutate'):
      AdWordsUtils.TransformJobOperationXsi(args[0])
    elif (self._service_name == 'UserListService' and
          method_name.lower() == 'mutate'):
      if isinstance(args[0], (list, tuple)):
        for operation in args[0]:
          if isinstance(operation, dict) and 'operand' in operation:
            AdWordsUtils.TransformUserListRuleOperands(operation['operand'])
    return args

  def _HandleLogsAndErrors(self, buf, start_time, stop_time, error=None):
    """Manage SOAP XML message.

    Args:
      buf: SoapBuffer SOAP buffer.
      start_time: str Time before service call was invoked.
      stop_time: str Time after service call was invoked.
      [optional]
      error: dict Error, if any.
    """
    if error is None:
      error = {}
    try:
      # Update the number of units and operations consumed by API call.
      if buf.GetCallUnits() and buf.GetCallOperations():
        self._config['units'][0] += int(buf.GetCallUnits())
        self._config['operations'][0] += int(buf.GetCallOperations())
        self._config['last_units'][0] = int(buf.GetCallUnits())
        self._config['last_operations'][0] = int(buf.GetCallOperations())

      handlers = self.__GetLogHandlers(buf)
      fault = super(GenericAdWordsService, self)._ManageSoap(
          buf, handlers, LIB_URL, start_time, stop_time, error)
      if fault:
        # Raise a specific error, subclass of AdWordsApiError.
        if 'detail' in fault and fault['detail']:
          if 'errors' in fault['detail']:
            error_type = fault['detail']['errors'][0]['type']
            if error_type in ERRORS: raise ERRORS[str(error_type)](fault)
        if isinstance(fault, basestring):
          raise AdWordsError(fault)
        elif isinstance(fault, dict):
          raise AdWordsApiError(fault)
    except AdWordsApiError, e:
      raise e
    except AdWordsError, e:
      raise e
    except Error, e:
      if error: e = error
      raise Error(e)

  def __GetLogHandlers(self, buf):
    """Gets a list of log handlers for the AdWords library.

    Args:
      buf: SoapBuffer SOAP buffer from which calls are retrieved for logging.

    Returns:
      list Log handlers for the AdWords library.
    """
    return [
        {
            'tag': 'xml_log',
            'name': 'soap_xml',
            'data': ''
        },
        {
            'tag': 'request_log',
            'name': 'request_info',
            'data': str('host=%s service=%s method=%s operator=%s '
                        'responseTime=%s operations=%s units=%s requestId=%s'
                        % (Utils.GetNetLocFromUrl(self._service_url),
                           self._service_name, buf.GetCallName(),
                           buf.GetOperatorName(), buf.GetCallResponseTime(),
                           buf.GetCallOperations(), buf.GetCallUnits(),
                           buf.GetCallRequestId()))
        },
        {
            'tag': '',
            'name': 'adwords_api_lib',
            'data': ''
        }
    ]


def _DetermineNamespacePrefix(url):
  """Returns the SOAP prefix to use for definitions within the given namespace.

  Args:
    url: string The URL of the namespace.

  Returns:
    string The SOAP namespace prefix to use for the given namespace.
  """
  return AdWordsUtils.ExtractGroupNameFromUrl(url) + ':'



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Downloads and patches SOAPpy into the current directory."""

__author__ = 'api.jdilallo@gmail.com (Joseph DiLallo)'

import os
import shutil
import subprocess
import tarfile
import tempfile
import urllib


# The DOWNLOAD_URL is the location where SOAPpy can be downloaded from.
DOWNLOAD_URL = ('http://sourceforge.net/projects/pywebsvcs/files/SOAP.py/'
                '0.12.0_rc1/SOAPpy-0.12.0.tar.gz/'
                'download?use_mirror=superb-dca2')
REMOVE_FILES = ['GSIServer.py']


def main():
  download_filename, _ = urllib.urlretrieve(DOWNLOAD_URL)
  tar_out_dir = tempfile.mkdtemp()
  downloaded_tar = tarfile.open(download_filename)
  downloaded_tar.extractall(tar_out_dir)

  for filename in os.listdir(os.path.join(tar_out_dir, 'SOAPpy-0.12.0',
                                          'SOAPpy')):
    abs_filename = os.path.join(tar_out_dir, 'SOAPpy-0.12.0', 'SOAPpy',
                                filename)
    if os.path.isdir(abs_filename):
      # If the destination exists (i.e. wstools), first remove it.
      if os.path.exists(os.path.join(os.curdir, filename)):
        shutil.rmtree(os.path.join(os.curdir, filename))
      shutil.copytree(abs_filename, os.path.join(os.curdir, filename))
    else:
      shutil.copy(abs_filename, os.curdir)

  code = subprocess.call(['patch', '-p0'],
                         stdin=open(os.path.join(os.curdir, 'SOAPpy.patch')))
  shutil.rmtree(tar_out_dir)
  for path in REMOVE_FILES:
    os.remove(path)
  if not code: 
    print 'Success! SOAPpy-0.12.0 has been downloaded and patched.'

  exit(code)


if __name__ == '__main__':
  main()



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handler functions for outgoing and incoming messages."""

__author__ = 'api.jdilallo@gmail.com (Joseph DiLallo)'

import types

from adspygoogle import SOAPpy
from adspygoogle.common import Utils
from adspygoogle.common.soappy import SoappyUtils


def UnpackResponseAsDict(response):
  """Unpack (recursively) SOAP data holder into a Python dict object.

  Args:
    response: instance SOAP data holder object.

  Returns:
    dict Unpacked SOAP data holder.
  """
  if (isinstance(response, types.InstanceType) and
      response.__dict__['_type'] == 'struct'):
    if not response.__dict__.keys(): return (response.__dict__,)
    dct = {}
    for key in response.__dict__:
      if key[0] == '_': continue
      value = response.__dict__.get(key)
      if key.find('.') > -1: key = key.replace('.', '_')
      data = UnpackResponseAsDict(value)
      dct[str(key)] = data
    return dct
  elif (isinstance(response, list) or
        (isinstance(response, types.InstanceType) and
         isinstance(response.__dict__['_type'], tuple)) or
        isinstance(response, SOAPpy.Types.typedArrayType)):
    lst = []
    for item in response:
      if item is None: continue
      lst.append(UnpackResponseAsDict(item))
    return lst
  elif isinstance(response, tuple) and len(response) == 6:
    return '%04d-%02d-%02dT%02d:%02d:%02d' % response
  else:
    if (isinstance(response, int) or isinstance(response, long) or
        isinstance(response, float)):
      return str(response)
    return response


def WrapInTuple(response):
  """Wraps the response in a tuple to support legacy behavior.

  Dictionaries and strings are placed in a tuple, while lists are unpacked into
  the tuple. None will not be wrapped in a tuple at all.

  Args:
    response: mixed The response from the API to wrap in a tuple. Could be a
              string, list, or dictionary depending on the operation.

  Returns:
    tuple: The given response wrapped in a tuple. Will be None if the input was
           None.
  """
  if response is None: return response
  if not isinstance(response, (list, tuple)): response = [response]
  return tuple(response)


def PackForSoappy(obj, xmlns, type_name, soappy_service, wrap_lists,
                  prefix_function):
  """Packs a given object into SOAPpy.Type objects for transport.

  Args:
    obj: mixed The python object to pack for SOAPpy transport. May be a string,
         list, or dictionary depending on what it represents.
    xmlns: string The namespace that the given object's type belongs to.
    type_name: string The name of the SOAP type this object represents.
    soappy_service: SOAPpy.WSDL.Proxy The SOAPpy service object encapsulating
                    this service/WSDL.
    wrap_lists: boolean Whether or not to wrap lists in an additional layer.
    prefix_function: callable Takes in an xml namespace and returns the prefix
                     to use to represent it.

  Returns:
    mixed The given object ready for SOAPpy transport. Depending on the input,
    this will be either a list, a SOAPpy.Types.structType, or a
    SOAPpy.Types.untypedType.
  """
  if isinstance(obj, dict):
    return _PackDictForSoappy(obj, xmlns, type_name, soappy_service, wrap_lists,
                              prefix_function)
  elif isinstance(obj, (list, tuple)):
    return _PackListForSoappy(obj, xmlns, type_name, soappy_service, wrap_lists,
                              prefix_function)
  elif isinstance(obj, str):
    return SOAPpy.Types.untypedType(Utils.HtmlEscape(obj).decode('utf-8'))
  elif isinstance(obj, unicode):
    return SOAPpy.Types.untypedType(Utils.HtmlEscape(obj))
  elif isinstance(obj, SOAPpy.Types.anyType) or obj is None:
    return obj
  else:
    return SOAPpy.Types.untypedType(obj)


def _PackDictForSoappy(obj, xmlns, type_name, soappy_service, wrap_lists,
                       prefix_function):
  """Packs a dictionary into a SOAPpy.Types.structType object for transport.

  Args:
    obj: dict The python object to pack for SOAPpy transport.
    xmlns: string The namespace that the given object's type belongs to.
    type_name: string The name of the SOAP type this object represents.
    soappy_service: SOAPpy.WSDL.Proxy The SOAPpy service object encapsulating
                    this service/WSDL.
    wrap_lists: boolean Whether or not to wrap lists in an additional layer.
    prefix_function: callable Takes in an xml namespace and returns the prefix
                     to use to represent it.

  Returns:
    SOAPpy.Types.structType The given dictionary ready for SOAPpy transport.
  """
  packed_data = {}
  obj_contained_type, type_key = SoappyUtils.GetExplicitType(
      obj, type_name, xmlns, soappy_service)
  if obj_contained_type:
    type_name = obj_contained_type

  for key in obj:
    if key == type_key or obj[key] is None:
      continue
    ns_prefix = prefix_function(SoappyUtils.GetComplexFieldNamespaceByFieldName(
        key, type_name, xmlns, soappy_service))
    key_type = SoappyUtils.GetComplexFieldTypeByFieldName(
        key, type_name, xmlns, soappy_service)
    packed_data[ns_prefix + key] = PackForSoappy(
        obj[key], key_type.getTargetNamespace(), key_type.getName(),
        soappy_service, wrap_lists, prefix_function)

  attrs = {(SOAPpy.NS.XSI3, 'type'): prefix_function(xmlns) + type_name}
  packed_object = SOAPpy.Types.structType(packed_data, typed=0, attrs=attrs)
  packed_object._typename = type_name
  packed_object._keyord = SoappyUtils.PruneKeyOrder(
      [param['name'] for param in SoappyUtils.GenKeyOrderAttrs(
          soappy_service, xmlns, type_name)], packed_object)
  return packed_object


def _PackListForSoappy(obj, xmlns, type_name, soappy_service, wrap_lists,
                       prefix_function, item_element_name='item'):
  """Packs an array input into a form ready for SOAPpy transport.

  Args:
    obj: list The python object to pack for SOAPpy transport.
    xmlns: string The namespace that the given object's type belongs to.
    type_name: string The name of the SOAP type this object represents.
    soappy_service: SOAPpy.WSDL.Proxy The SOAPpy service object encapsulating
                    this service/WSDL.
    wrap_lists: boolean Whether or not to wrap lists in an additional layer.
    prefix_function: callable Takes in an xml namespace and returns the prefix
                     to use to represent it.
    [optional]
    item_element_name: string The XML element name to wrap items of the list in.
                       Only used if wrap_lists is set to True. Defaults to
                       'item'.

  Returns:
    mixed The given object ready for SOAPpy transport. If wrap_lists is set to
    False, this will return a list. If wrap_lists is set to True, this will
    return a SOAPpy.Types.structType.
  """
  if wrap_lists:
    new_list = SOAPpy.Types.structType(typed=0)
    for item in obj:
      item_type = SoappyUtils.GetArrayItemTypeName(type_name, xmlns,
                                                   soappy_service)
      new_list._addItem(item_element_name, PackForSoappy(
          item, xmlns, item_type, soappy_service, wrap_lists,
          prefix_function))
    return new_list
  else:
    new_list = []
    for item in obj:
      item_type = SoappyUtils.GetArrayItemTypeName(type_name, xmlns,
                                                   soappy_service)
      new_list.append(PackForSoappy(item, xmlns, item_type, soappy_service,
                                    wrap_lists, prefix_function))
    return new_list


def RestoreListTypeWithSoappy(response, service, operation_return_types):
  """Restores list types within given response which were overwritten by SOAPpy.

  Args:
    response: mixed A string, list, or dict of response data.
    service: SOAPpy.WSDL.Proxy A SOAPpy service object encapsulating the WSDL
             definitions.
    operation_return_types: list Data types this operation returns, in
                            order.

  Returns:
    String, list, or dict: The given response with nested items corrected if
    necessary. Should always be the same data type as the input response.
  """
  if not operation_return_types:
    # If the method doesn't return anything, return None. SOAPpy returns {}.
    return None
  if len(operation_return_types) > 1:
    new_response = []
    for i in len(response):
      new_response.append(RestoreListTypeWithSoappy(response[i], service,
                                                    operation_return_types[i]))
    return new_response

  return_ns, return_type_name, return_max_occurs = operation_return_types[0]

  if not return_max_occurs.isdigit() or int(return_max_occurs) > 1:
    if not response:
      return []
    elif not isinstance(response, list):
      return [_RestoreListTypesForResponse(
          response, return_type_name, return_ns, service)]
    else:
      return _RestoreListTypesForResponse(
          response, return_type_name, return_ns, service)
  else:
    return _RestoreListTypesForResponse(
        response, return_type_name, return_ns, service)


def _RestoreListTypesForResponse(response, type_name, ns, service):
  """Restores list types for an individual response object.

  Args:
    response: mixed An individual object returned by the webservice. May be a
              dictionary, list, or string depending on what it represents.
    type_name: string The type name defined in the WSDL.
    ns: string The namespace the given type belongs to.
    service: SOAPpy.WSDL.Proxy An object encapsulating a SOAP service.

  Returns:
    obj The response in its proper format. May be a dictionary, list, or string
    depending on what was input. Not guaranteed to output the same data type
    that was input - may output a list instead.
  """
  if isinstance(response, dict):
    if not response: return response
    for key in response:
      if key.endswith("_Type"): type_name = response[key]
    parameters = SoappyUtils.GenKeyOrderAttrs(service, ns, type_name)
    for param, param_type, param_max_occurs in [
        (param['name'], param['type'], param['maxOccurs'])
        for param in parameters]:
      if not param in response:
        continue
      value = response[param]
      if ((SoappyUtils.IsAnArrayType(
          param_type.getName(), param_type.getTargetNamespace(), service) or
           (not param_max_occurs.isdigit() or int(param_max_occurs) > 1)) and
          not isinstance(response[param], list)):
        value = [value]
      response[param] = _RestoreListTypesForResponse(
          value, param_type.getName(), param_type.getTargetNamespace(), service)
    return response
  elif isinstance(response, list):
    return [_RestoreListTypesForResponse(item, SoappyUtils.GetArrayItemTypeName(
        type_name, ns, service), ns, service) for item in response if item]
  else:
    return response



#!/usr/bin/python
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for handling errors."""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

from adspygoogle.common.Errors import DetailError
from adspygoogle.common.Errors import Error


class AdWordsError(Error):

  """Implements AdWordsError.

  Responsible for handling error.
  """
  pass


class AdWordsDetailError(DetailError):

  """Implements AdWordsDetailError.

  Responsible for handling detailed ApiException error.
  """

  def __init__(self, **error):
    DetailError.__init__(self)
    self.__error = error
    for key in self.__error:
      self.__dict__.__setitem__(key, self.__error[key])

  def __call__(self):
    return (self.__error,)


class AdWordsApiError(AdWordsError):

  """Implements AdWordsApiError.

  Responsible for handling API exception.
  """

  def __init__(self, fault):
    (self.__fault, self.fault_code, self.fault_string) = (fault, '', '')
    if 'faultcode' in fault: self.fault_code = fault['faultcode']
    if 'faultstring' in fault: self.fault_string = fault['faultstring']

    (self.code, error_msg, self.trigger, self.errors) = (-1, '', '', [])
    if 'detail' in fault and fault['detail']:
      if 'code' in fault['detail']: self.code = int(fault['detail']['code'])
      if 'trigger' in fault['detail']: self.trigger = fault['detail']['trigger']
      if 'message' in fault['detail']: error_msg = fault['detail']['message']
    if not error_msg: error_msg = self.fault_string

    errors = [None]
    if 'detail' in fault and fault['detail'] and 'errors' in fault['detail']:
      errors = fault['detail']['errors']
    elif 'detail' not in fault or not fault['detail']:
      errors[0] = {}
    else:
      errors[0] = fault['detail']
    if isinstance(errors, list):
      for error in errors:
        # Keys need to be of type str not unicode.
        error_dct = dict([(str(key), value) for key, value in error.items()])
        self.errors.append(AdWordsDetailError(**error_dct))
    else:
      # Keys need to be of type str not unicode.
      error_dct = dict([(str(key), value) for key, value in errors.items()])
      self.errors.append(AdWordsDetailError(**error_dct))
    AdWordsError.__init__(self, error_msg)

  def __call__(self):
    return (self.__fault,)


class AdWordsRequestError(AdWordsApiError):

  """Implements AdWordsRequestError.

  Responsible for handling request error."""

  pass


class AdWordsGoogleInternalError(AdWordsApiError):

  """Implements AdWordsGoogleInternalError.

  Responsible for handling Google internal error.
  """

  pass


class AdWordsAuthenticationError(AdWordsApiError):

  """Implements AdWordsAuthenticationError.

  Responsible for handling authentication error.
  """

  pass


class AdWordsAccountError(AdWordsApiError):

  """Implements AdWordsAccountError.

  Responsible for handling account error.
  """

  pass


class AdWordsWebpageError(AdWordsApiError):

  """Implements AdWordsWebpageError.

  Responsible for handling webpage error.
  """

  pass


class AdWordsBillingError(AdWordsApiError):

  """Implements AdWordsBillingError.

  Responsible for handling billing error.
  """

  pass


class AdWordsReportError(AdWordsError):

  def __init__(self, http_code, error_type, trigger, field_path):
    self.http_code = http_code
    self.type = error_type
    self.trigger = trigger
    self.field_path = field_path
    message = ('HTTP code: %s, type: \'%s\', trigger: \'%s\', '
               'field path: \'%s\'' %
               (http_code, error_type, trigger, field_path))
    super(AdWordsReportError, self).__init__(message)


# Map error codes and types to their corresponding classes.
ERRORS = {}
for e_type in ('AdError', 'AdExtensionError', 'AdExtensionOverrideError',
               'AdGroupAdError', 'AdGroupCriterionError',
               'AdGroupServiceError', 'AdParamError', 'AdParamPolicyError',
               'AlertError', 'ApiError', 'ApiUsageError', 'AudioError',
               'BidLandscapeServiceError', 'BiddingError',
               'BiddingTransitionError', 'BudgetError', 'BulkMutateJobError',
               'CampaignAdExtensionError',
               'CampaignCriterionError', 'CampaignError',
               'CollectionSizeError', 'ConversionTrackingError',
               'CriterionError', 'CriterionPolicyError', 'CurrencyCodeError',
               'CustomerSyncError', 'DataError', 'DateError', 'DistinctError',
               'ExperimentServiceError', 'GeoLocationError', 'IdError',
               'ImageError', 'JobError', 'MatchesRegexError', 'MediaError',
               'NewEntityCreationError', 'NotEmptyError', 'NullError',
               'OperatorError', 'OpportunityError', 'PagingError',
               'PolicyViolationError', 'RangeError', 'ReadOnlyError',
               'RegionCodeError', 'RejectedError', 'ReportDefinitionError',
               'RequestError', 'RequiredError', 'SelectorError',
               'ServicedAccountError', 'SettingError', 'SizeLimitError',
               'StatsQueryError', 'StringLengthError', 'TargetError',
               'TargetingIdeaError', 'TrafficEstimatorError', 'UserListError',
               'VideoError'):
  ERRORS[e_type] = AdWordsRequestError
for e_type in ('InternalApiError', 'DatabaseError', 'QuotaCheckError',
               'QuotaError', 'QuotaExceededError', 'RateExceededError'):
  ERRORS[e_type] = AdWordsGoogleInternalError
for e_type in ('ClientTermsError', 'NotWhitelistedError'):
  ERRORS[e_type] = AdWordsAccountError
for e_type in ('AuthenticationError', 'AuthorizationError'):
  ERRORS[e_type] = AdWordsAuthenticationError



#!/usr/bin/python
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains common constants for PyPI release scripts."""
__author__ = 'api.msaniscalchi@gmail.com (Mark Saniscalchi)'

import warnings
from adspygoogle.common import GenerateLibSig

try:
  from adspygoogle.adwords.AdWordsClient import AdWordsClient
except ImportError, e:
  warnings.warn('Can\'t import AdWordsClient: %s' % e)

try:
  from adspygoogle.dfa.DfaClient import DfaClient
except ImportError, e:
  warnings.warn('Can\'t import DfaClient: %s' % e)

try:
  from adspygoogle.dfp.DfpClient import DfpClient
except ImportError, e:
  warnings.warn('Can\'t import DfpClient: %s' % e)

LIB_PYPI_VERSION = '1.1.6'
LIB_PYPI_NAME = 'Google Ads Python Client Library'
LIB_PYPI_SHORT_NAME = 'adspygoogle'
LIB_PYPI_URL = 'https://github.com/googleads/googleads-python-lib'
LIB_PYPI_AUTHOR = 'Joseph DiLallo'
LIB_PYPI_AUTHOR_EMAIL = 'api.jdilallo@gmail.com'
LIB_PYPI_SIG = GenerateLibSig(LIB_PYPI_SHORT_NAME, LIB_PYPI_VERSION)



#!/usr/bin/python
# -*- coding: UTF-8 -*-
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script to fetch codes used throughout the AdWords API web services.

See http://code.google.com/apis/adwords/docs/developer/adwords_api_codes.html.
"""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

import os
import re
import sys
sys.path.insert(0, os.path.join('..', '..', '..'))
import urllib


from adspygoogle.common import Utils


URL = 'http://code.google.com/apis/adwords/docs/developer'
LOC = os.path.join('..', 'aw_api', 'data')
DATA_MAP = [
  {'csv': 'categories',
   'url': '/'.join([URL, 'adwords_api_categories.html']),
   're': ('<li><b>/(.*?)</b></li>|<li><span class="categorypath">(.*?)</span>'
          '<b>(.*?)</b></li>'),
   'cols': ['category', 'path']
  },
  {'csv': 'countries',
   'url': '/'.join([URL, 'adwords_api_countries.html']),
   're': '<tr><td>(.*?)</td><td>(.*?)</td></tr>',
   'cols': ['country', 'code']
  },
  {'csv': 'currencies',
   'url': '/'.join([URL, 'adwords_api_currency.html']),
   're': '<tr><td>(.*?)</td><td>(.*?)</td></tr>',
   'cols': ['code', 'currency']
  },
  {'csv': 'error_codes',
   'url': '/'.join([URL, 'adwords_api_error_codes.html']),
   're': ('<tr id=".*?" class="ShowHide"><td id=".*?"><code><span class="">'
          '(.*?)</span></code></td><td><span class=""> (.*?)</span></td></tr>'),
   'cols': ['code', 'message']
  },
  {'csv': 'languages',
   'url': '/'.join([URL, 'adwords_api_languages.html']),
   're': '<tr><td>(.*?)</td><td>(.*?)</td><td>(.*?)</td></tr>',
   'cols': ['name', 'target', 'display']
  },
  {'csv': 'timezones',
   'url': '/'.join([URL, 'adwords_api_timezones.html']),
   're': '<tr><td>(.*?)</td></tr>',
   'cols': ['timezone']
  },
  {'csv': 'us_cities',
   'url': '/'.join([URL, 'adwords_api_us_cities.html']),
   're': '<h2 id=".*?">(.*?)</h2><table.*?>(.*?)</table>',
   'cols': ['state', 'code']
  },
  {'csv': 'us_metros',
   'url': '/'.join([URL, 'adwords_api_us_metros.html']),
   're': '<table class="codes" summary="">.*?<h2>(.*?)</h2>(.*?)</table>',
   'cols': ['state', 'metro', 'code']
  },
  {'csv': 'world_cities',
   'url': '/'.join([URL, 'adwords_api_cities.html']),
   're': ('<h2 id=".*?">((?:\w+|\w+\s\w+))</h2><table class="codes" summary="" '
          'style="width: 100%"><tr>(.*?)(?:</ul></td>|)</tr></table>'),
   'cols': ['country', 'code']
  },
  {'csv': 'world_regions',
   'url': '/'.join([URL, 'adwords_api_regions.html']),
   're': ('<h2 id=".*?">((?:\w+|\w+\s\w+))</h2><table class="codes" '
          'summary="codes">(.*?)</table>'),
   'cols': ['country', 'code', 'region']
  }
]


if os.path.exists(os.path.abspath(LOC)):
  for f_name in os.listdir(os.path.abspath(LOC)):
    f_path = os.path.abspath(os.path.join(LOC, f_name))
    if f_name.split('.')[-1] == 'csv':
      os.unlink(f_path)
else:
  os.mkdir(os.path.abspath(LOC))

print 'Fetching codes ...'
for item in DATA_MAP:
  data = ''.join(urllib.urlopen(item['url']).read().split('\n'))

  # Remove weird unicode characters.
  for weird_char in ['\xa0', '\xc2']:
    if data.find(weird_char) > -1:
      data = data.replace(weird_char, '')

  pattern = re.compile(item['re'])
  groups = pattern.findall(data)
  lines = []
  for group in groups:
    if item['csv'] in ('categories',):
      group = [x for x in group if x != '']
      if len(group) == 1:
        category = ''.join(group)
        continue
      else:
        path = ''.join(group)
      lines.append('%s,%s'
                   % (Utils.CsvEscape(str(Utils.HtmlUnescape(category))),
                      Utils.CsvEscape(str(Utils.HtmlUnescape(path)))))
    elif item['csv'] in ('countries', 'currencies'):
      lines.append('%s,%s'
                   % (Utils.CsvEscape(str(Utils.HtmlUnescape(group[0]))),
                      Utils.CsvEscape(str(Utils.HtmlUnescape(group[1])))))
    elif item['csv'] in ('error_codes',):
      pattern = re.compile('<.*?>')
      message = list(group)[1]
      message = pattern.sub('', message)
      lines.append('%s,%s'
                   % (Utils.CsvEscape(str(Utils.HtmlUnescape(group[0]))),
                      Utils.CsvEscape(str(Utils.HtmlUnescape(message)))))
    elif item['csv'] in ('languages',):
      # Convert '-' into ''.
      new_group = []
      for sub_item in list(group):
        if sub_item == '-':
          new_group.append('')
        else:
          new_group.append(sub_item)
      lines.append('%s,%s,%s'
                   % (Utils.CsvEscape(str(Utils.HtmlUnescape(new_group[0]))),
                      Utils.CsvEscape(str(Utils.HtmlUnescape(new_group[1]))),
                      Utils.CsvEscape(str(Utils.HtmlUnescape(new_group[2])))))
    elif item['csv'] in ('timezones',):
      lines.append('%s' % str(Utils.HtmlUnescape(group)))
    elif item['csv'] in ('us_cities',):
      if group[0] != 'States':
        pattern = re.compile('<li>(.*?)</li>')
        sub_groups = pattern.findall(group[1])
        for sub_group in sub_groups:
          lines.append('%s,%s'
                       % (Utils.CsvEscape(str(Utils.HtmlUnescape(group[0]))),
                          Utils.CsvEscape(str(Utils.HtmlUnescape(sub_group)))))
    elif item['csv'] in ('us_metros',):
      pattern = re.compile('<tr><td>(.*?)</td><td>(.*?)</td></tr>')
      sub_groups = pattern.findall(group[1])
      for sub_group in sub_groups:
        lines.append('%s,%s,%s'
                     % (Utils.CsvEscape(str(Utils.HtmlUnescape(group[0]))),
                        Utils.CsvEscape(str(Utils.HtmlUnescape(sub_group[0]))),
                        Utils.CsvEscape(str(Utils.HtmlUnescape(sub_group[1])))))
    elif item['csv'] in ('world_cities',):
      pattern = re.compile('<li>(.*?)</li>')
      sub_groups = pattern.findall(group[1])
      for sub_group in sub_groups:
        lines.append('%s,%s'
                     % (Utils.CsvEscape(str(Utils.HtmlUnescape(group[0]))),
                        Utils.CsvEscape(str(Utils.HtmlUnescape(sub_group)))))
    elif item['csv'] in ('world_regions',):
      pattern = re.compile('<tr><td>(.*?)</td><td>(.*?)</td></tr>')
      sub_groups = pattern.findall(group[1])
      for sub_group in sub_groups:
        lines.append('%s,%s,%s'
                     % (Utils.CsvEscape(str(Utils.HtmlUnescape(group[0]))),
                        Utils.CsvEscape(str(Utils.HtmlUnescape(sub_group[0]))),
                        Utils.CsvEscape(str(Utils.HtmlUnescape(sub_group[1])))))

  print '  [+] %s.csv' % item['csv']
  fh = open('%s.csv' % os.path.abspath(os.path.join(LOC, item['csv'])), 'w')
  try:
    lines.insert(0, ','.join(item['cols']))
    for line in lines:
      fh.write('%s\n' % line)
  finally:
    fh.close()

print '... done.'
